/*
 * Arduino toolbox for Linux
 * copyright U. Raich, Dec 2018
 */
#include <sys/types.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <stdio.h>

#include "arduino_serial.h"

#define DEBUG 1
struct termios oldtio,newtio;
int noOfCharsRead,noOfCharsWritten;
int correctHandle;

void open_serial(int *handle,int *port, int *baudrate, int *errCode){
  int fd;
  tcflag_t br;
  char modemDevice[15];

  if ((*port < 0) || (*port > 3)) {
    *errCode = ARDUINO_ILLEGAL_PORT;
    return;
  }
  sprintf(modemDevice,"/dev/ttyUSB%d",*port);
#ifdef DEBUG
  printf("open_serial: modem device: %s\n",modemDevice);
#endif

  switch (*baudrate) {
  case 9600:
    br = B9600;
    break;
  case 19200:
    br = B19200;
    break;
  case 38400:
    br = B38400;
    break;
  case 57600:
    br = B57600;
    break;
  case 115200:
    printf("Setting baudrate to 115200 baud\n");
    br = B115200;
    break;
  default:
    printf("open_serial: Illegal baud rate\n");
    *errCode = ARDUINO_ILLEGAL_BAUDRATE;
    return;
  }
  fd = open(modemDevice, O_RDWR | O_NOCTTY ); 
  if (fd <0) {
    printf("open_serial: error when trying to open serial line %s\n",
        modemDevice);
    *errCode = ARDUINO_BAD_OPEN;
    return;
  }
  else {
    *handle = fd;
    correctHandle=fd;
    sleep(2);
  }
    
#ifdef DEBUG
  printf("open_serial: Serial port successfully opened with handle: %d\n",fd);
#endif

  /*
    set the baud rate
  */
  
  // Initialize file descriptor sets
  fd_set read_fds, write_fds, except_fds;
  FD_ZERO(&read_fds);
  FD_ZERO(&write_fds);
  FD_ZERO(&except_fds);
  FD_SET(fd, &read_fds);

  tcgetattr(fd,&oldtio); /* save current port settings */

  bzero(&newtio, sizeof(newtio));
  newtio.c_cflag = br | CRTSCTS | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;
  
  /* set input mode (non-canonical, no echo,...) */
  newtio.c_lflag = 0;
   
  newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
  newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 char received */
  
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW,&newtio);
  noOfCharsRead = noOfCharsWritten = 0;
  sleep(2);
  *errCode = ARDUINO_SUCCESS;
  return;
}

void close_serial(int *handle, int *errCode) {
  if (*handle != correctHandle)
    printf("Warning: the handle is incorrect: should be: %d but is %d\n",
        correctHandle,*handle);
#ifdef DEBUG
  printf("close_serial: resetting serial line pars and closing\n");
#endif
  tcsetattr(*handle,TCSANOW,&oldtio);
  close(*handle);
  return;
}

void write_serial(int *handle, char *str, int *size, int *errCode) {
  int i;
  if (*handle != correctHandle)
    printf("Warning: the handle is incorrect: should be: %d but is %d\n",
        correctHandle,*handle);
  if ((noOfCharsWritten = write(*handle, str, *size)) != *size)
    *errCode = ARDUINO_WRITE_ERROR;
  else {
#ifdef DEBUG
    printf("write_serial: wrote %d characters to serial port\n",*size);
    printf("The text was: ");
    for (i=0; i<*size; i++)
      printf("%c",str[i]);
    printf("\n");
#endif
  }
  return;
}

void read_serial(int *handle, char *buf, int *size, int *errCode) {

  int fd=*handle;
  struct timeval timeout;
  timeout.tv_sec =  5;
  timeout.tv_usec = 0;
  
  if (*handle != correctHandle)
    printf("Warning: the handle is incorrect: should be: %d but is %d\n",
	   correctHandle,*handle); 
  // Initialize file descriptor sets
  fd_set read_fds, write_fds, except_fds;
  FD_ZERO(&read_fds);
  FD_ZERO(&write_fds);
  FD_ZERO(&except_fds);
  FD_SET(fd, &read_fds);

  if(select(fd+1,&read_fds,&write_fds,&except_fds,&timeout) == 1) {
    if (read(fd,buf,*size) != *size) {
      *errCode = ARDUINO_READ_ERROR;
      return;
    }
    buf[*size]='\0';
    printf("read: %s\n",buf);
    *errCode = ARDUINO_SUCCESS;
  }
  else {
    printf("Timeout on reading\n");
    *errCode = ARDUINO_READ_TIMEOUT;
  }
  return;
}

void status_serial(int *handle,char *errCode, int *nbread, int *nbwrite) {
  *nbread = noOfCharsRead;
  *nbwrite = noOfCharsWritten;
  *errCode = ARDUINO_SUCCESS;
}
